# SPDX-License-Identifier: Apache-2.0
import functools
import itertools
import math
import os
import queue
import signal
import threading
import time
import traceback
from typing import Any, Callable, Optional, Tuple, TypeVar, Union

import jax
# ======================================================================================
# Imports for DisaggEngineCoreProc (the vLLM adapter)
# ======================================================================================
from vllm.config import VllmConfig
from vllm.logger import init_logger
from vllm.tasks import POOLING_TASKS, SupportedTask
from vllm.v1.core.kv_cache_utils import (get_request_block_hasher,
                                         init_none_hash)
from vllm.v1.engine import (EngineCoreOutputs, EngineCoreRequest,
                            EngineCoreRequestType, UtilityOutput,
                            UtilityResult)
from vllm.v1.engine.core import EngineCore as vLLMEngineCore
from vllm.v1.engine.core import EngineCoreProc as vLLMEngineCoreProc
from vllm.v1.executor.abstract import Executor
from vllm.v1.request import Request, RequestStatus

from tpu_inference import utils as common_utils
from tpu_inference.core import disagg_executor, disagg_utils
from tpu_inference.runner.tpu_runner import AsyncTPUModelRunnerOutput
# ======================================================================================
# Imports for _DisaggOrchestrator (decoupled from vLLM)
# ======================================================================================
from tpu_inference.runner.utils import LatencyTracker

# This file contains two classes:
# 1. _DisaggOrchestrator: The clean, decoupled core orchestration logic.
# 2. DisaggEngineCoreProc: The vLLM-facing adapter that handles process management.

logger = init_logger(__name__)

POLLING_TIMEOUT_S = 2.5
HANDSHAKE_TIMEOUT_MINS = 5

_R = TypeVar('_R')  # Return type for collective_rpc

# ======================================================================================
# Class 1: The Decoupled Orchestrator
# ======================================================================================


class JetThread(threading.Thread):
    """Thread that kills the program if it fails.

    If a driver thread goes down, we can't operate.
    """

    def run(self):
        try:
            super().run()
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(f"Thread {self.name} encountered an error: {e}")
            traceback.print_exc()
            os.kill(os.getpid(), signal.SIGKILL)


class _DisaggOrchestrator:
    """Contains the core orchestration logic, decoupled from vLLM."""

    def __init__(
        self,
        config: VllmConfig,
        output_queue: queue.Queue,
        prefill_engines: list[vLLMEngineCore],
        decode_engines: list[vLLMEngineCore],
        prefill_slice_sizes: tuple[int, ...],
        decode_slice_sizes: tuple[int, ...],
    ):
        self._config = config
        self._output_queue = output_queue
        self._prefill_engines = prefill_engines
        self._decode_engines = decode_engines

        # Keep track of active requests.
        self._requests: dict[str, Request] = {}

        # Hack device config to pass in the subslice of TPUs.
        slice_sizes = list(prefill_slice_sizes)
        slice_sizes.extend(decode_slice_sizes)

        self._transfer_backlogs = [
            queue.Queue(4) for i in range(len(self._prefill_engines))
        ]

        self._decode_backlogs = {}
        for idx, engine in enumerate(self._decode_engines):
            # Determine the decode backlog len by remaning hbm dividing max kv cache size of a single request
            runner = engine.model_executor.driver_worker.model_runner
            hbm_usage = common_utils.hbm_usage_bytes(
                engine.model_executor.driver_worker.devices)
            if not hbm_usage:
                self._decode_backlogs[idx] = queue.Queue(
                    self._config.scheduler_config.max_num_seqs)
                continue
            hbm_free = [limit - used for used, limit in hbm_usage]
            max_kv_bytes = len(runner.kv_caches) * (
                runner.max_model_len // runner.cache_config.block_size) * (
                    runner.kv_caches[0][0].nbytes) // len(hbm_free)
            max_queue_len = min(hbm_free[0] // max_kv_bytes,
                                self._config.scheduler_config.max_num_seqs)
            logger.debug(
                f"max kv bytes: {max_kv_bytes}, max_queue_len {max_queue_len}")
            self._decode_backlogs[idx] = queue.Queue(max_queue_len)

        self._prefill_threads = [
            JetThread(
                target=functools.partial(self._prefill, idx),
                name=f"prefill-{idx}",
                daemon=True,
            ) for idx in range(len(self._prefill_engines))
        ]
        self._transfer_threads = [
            JetThread(
                target=functools.partial(
                    self._transfer,
                    idx,
                ),
                name=f"transfer-{idx}",
                daemon=True,
            ) for idx in range(len(self._prefill_engines))
        ]
        self._decode_threads = [
            JetThread(
                target=functools.partial(
                    self._decode,
                    idx,
                ),
                name=f"decode-{idx}",
                daemon=True,
            ) for idx in range(len(self._decode_engines))
        ]
        self._all_threads = list(
            itertools.chain(
                self._prefill_threads,
                self._transfer_threads,
                self._decode_threads,
            ))
        self.live = True
        # Start all threads
        for t in self._all_threads:
            t.start()

    def add_request(self, request: Request):
        """
        Adds a new request to the orchestrator.

        This is the main entry point for new work. It stores the request for
        internal state tracking and hands it off to the first stage of the
        processing pipeline (the prefill scheduler).
        """
        # Hand off the request to the prefill scheduler to be batched for execution.
        self._prefill_engines[0].scheduler.add_request(request)

        # Add to internal state for tracking by other threads.
        # The key is the request_id, the value is the request object.
        self._requests[request.request_id] = request

    def _prefill(self, idx: int):
        prefill_engine = self._prefill_engines[idx]
        transfer_backlog = self._transfer_backlogs[idx]

        while self.live:
            if not prefill_engine.scheduler.has_requests():
                time.sleep(0.05)
                continue

            scheduler_output = prefill_engine.scheduler.schedule()
            with LatencyTracker(f"prefill-{idx}"):
                future = prefill_engine.model_executor.execute_model(
                    scheduler_output, non_block=True)
                grammar_output = prefill_engine.scheduler.get_grammar_bitmask(
                    scheduler_output)
                with prefill_engine.log_error_detail(scheduler_output):
                    model_output = future.result()
                    if model_output is None:
                        model_output = prefill_engine.model_executor.sample_tokens(
                            grammar_output)
                    if isinstance(model_output, AsyncTPUModelRunnerOutput):
                        model_output = model_output.get_output()

            if scheduler_output.total_num_scheduled_tokens > 0:
                logger.debug(f"Prefill result: {model_output}")

                kv_cache_map: dict[str, Tuple(list[jax.Array], list[Any])] = {}
                for req_id, idx in model_output.req_id_to_index.items():
                    if len(model_output.sampled_token_ids[idx]) > 0:
                        request = self._requests[req_id]
                        block_ids = (prefill_engine.scheduler.kv_cache_manager.
                                     get_block_ids(req_id))
                        # Assume one KV cache group for now.
                        kv_cache_map[req_id] = (
                            prefill_engine.model_executor.driver_worker.
                            model_runner.get_kv_cache_for_block_ids(
                                block_ids[0]), request.block_hashes)
                        logger.debug(f"prefill done: for {req_id}")
                transfer_backlog.put(kv_cache_map, block=True)

                # tweak model_output to let the scheduler know kv_transfer is done for requests, so they can be freed.
                engine_core_outputs = prefill_engine.scheduler.update_from_output(
                    scheduler_output, model_output)  # type: ignore

                for req_id, idx in model_output.req_id_to_index.items():
                    if len(model_output.sampled_token_ids[idx]) > 0:
                        request = self._requests[req_id]
                        logger.debug(
                            f"request block_hashes at prefill: {request.block_hashes}"
                        )
                        logger.debug(
                            f"request-{req_id}: tokens={request.all_token_ids} after prefill"
                        )
                        # Remove request from the prefill engine.
                        if req_id in prefill_engine.scheduler.requests:
                            request = prefill_engine.scheduler.requests[req_id]
                            prefill_engine.scheduler.running.remove(request)
                            prefill_engine.scheduler.encoder_cache_manager.free(
                                request)

                            prefill_engine.scheduler.kv_cache_manager.free(
                                request)

                            prefill_engine.scheduler.requests.pop(req_id)

                for output in (engine_core_outputs.items()
                               if engine_core_outputs else ()):
                    self._output_queue.put_nowait(output)

    def _transfer(self, idx: int):
        """Transfers the kv cache on an active request to the least full
    decode backlog."""
        transfer_backlog = self._transfer_backlogs[idx]
        while self.live:
            # The transfer thread can just sleep until it has work to do.
            kv_cachce_map = transfer_backlog.get(block=True)
            if kv_cachce_map is None:
                break

            logger.debug(
                f"transfer-{idx}: KV Cache items received: {kv_cachce_map.keys()}"
            )

            push_targets = []
            for req_id, (kv_cache, block_hashes) in kv_cachce_map.items():
                target_idx = -1
                cnt = 9999999
                for i, e in enumerate(self._decode_engines):
                    req_cnt = sum(e.scheduler.get_request_counts())
                    if req_cnt < cnt:
                        cnt = req_cnt
                        target_idx = i

                # Only transfer the KVCache for the disaggregated serving.
                with LatencyTracker("KVCacheTransfer"):
                    kv_cache = self._decode_engines[
                        target_idx].model_executor.driver_worker.model_runner.transfer_kv_cache(
                            kv_cache)

                # TODO(fhzhang): Now how do we get the kv cache to the decode engine?
                prefill_output = {
                    "cache": kv_cache,
                    "req_id": req_id,
                    "block_hashes": block_hashes,
                }
                push_targets.append((target_idx, prefill_output))

            for target_idx, prefill_output in push_targets:
                self._decode_backlogs[target_idx].put(prefill_output,
                                                      block=True)
                logger.debug(
                    "Successfully transferred prefill request %s "
                    "from prefill engine %d to decode engine %d. decode backlog len %d",
                    prefill_output["req_id"],
                    idx,
                    target_idx,
                    self._decode_backlogs[target_idx].qsize(),
                )

    def _decode(self, idx: int):
        decode_engine = self._decode_engines[idx]
        decode_backlog = self._decode_backlogs[idx]

        while self.live:
            block = not decode_engine.scheduler.has_requests()

            while True:
                # We need to check input batch as well as the request completion is delayed
                # from scheduler to the runner.
                if (sum(decode_engine.scheduler.get_request_counts())
                        >= self._config.scheduler_config.max_num_seqs
                        or decode_engine.model_executor.driver_worker.
                        model_runner.input_batch.num_reqs
                        >= self._config.scheduler_config.max_num_seqs):
                    break

                try:
                    prefill_output = decode_backlog.get(block=block,
                                                        timeout=1.0)
                except queue.Empty:
                    if block:
                        continue
                    break

                if prefill_output is None:
                    logger.debug(
                        f"decode-{idx} Empty output, and we are idle, exiting..."
                    )
                    break

                # We got a request, set block to False
                block = False

                # Insert the request to the decoder.
                req_id = prefill_output["req_id"]
                vllm_request = self._requests[req_id]
                # Caching num_computed_tokens. The tokens in kv manager allocate blocks
                # is computed as num_computed_tokens + num_new_tokens, so without caching
                # the token number would double.
                prompt_tokens = vllm_request.num_computed_tokens
                vllm_request.num_computed_tokens = 0
                kv_cache = prefill_output["cache"]

                kv_cache_manager = decode_engine.scheduler.kv_cache_manager
                kv_cache_manager.allocate_slots(
                    vllm_request,
                    prompt_tokens,
                )
                vllm_request.num_computed_tokens = prompt_tokens
                new_block_ids = kv_cache_manager.get_block_ids(req_id)
                logger.debug(
                    f"inserting {req_id} new_block_ids {new_block_ids}")
                if len(new_block_ids[0]) != math.ceil(
                        prompt_tokens / self._config.cache_config.block_size):
                    logger.warning("Running out of blocks in decode engine! ")
                    break

                decode_engine.model_executor.driver_worker.model_runner.insert_request_with_kv_cache(
                    vllm_request, kv_cache, new_block_ids)

                vllm_request.status = RequestStatus.RUNNING
                block_hashes = prefill_output["block_hashes"]
                vllm_request.block_hashes = block_hashes
                decode_engine.scheduler.running.append(vllm_request)
                decode_engine.scheduler.requests[req_id] = vllm_request

                self._requests.pop(req_id)

            scheduler_output = decode_engine.scheduler.schedule()

            logger.debug(f'''decode-{idx}: scheduler_output -
                {scheduler_output.scheduled_cached_reqs.num_computed_tokens},
                new block ids - {scheduler_output.scheduled_cached_reqs.new_block_ids}'''
                         )

            with LatencyTracker(f"decode-{idx}"):
                future = decode_engine.model_executor.execute_model(
                    scheduler_output, non_block=True)
                grammar_output = decode_engine.scheduler.get_grammar_bitmask(
                    scheduler_output)
                with decode_engine.log_error_detail(scheduler_output):
                    model_output = future.result()
                    if model_output is None:
                        model_output = decode_engine.model_executor.sample_tokens(
                            grammar_output)
                    if isinstance(model_output, AsyncTPUModelRunnerOutput):
                        model_output = model_output.get_output()

            if scheduler_output.total_num_scheduled_tokens > 0:
                logger.debug(f"Decode result: {model_output}")

                engine_core_outputs = decode_engine.scheduler.update_from_output(
                    scheduler_output, model_output)  # type: ignore
                for output in (engine_core_outputs.items()
                               if engine_core_outputs else ()):
                    self._output_queue.put_nowait(output)

    def shutdown(self):
        for e in self._prefill_engines:
            e.shutdown()
        for e in self._decode_engines:
            e.shutdown()


# ======================================================================================
# Class 2: The vLLM-Facing Adapter
# ======================================================================================


def _create_engine_cores(
    slice_sizes: tuple[int, ...],
    vllm_config: VllmConfig,
    log_stats: bool,
    executor_fail_callback: Optional[Callable] = None,
) -> list[vLLMEngineCore]:
    engine_cores = []
    for _ in slice_sizes:
        engine_core = vLLMEngineCore(
            vllm_config,
            disagg_executor.DisaggExecutor,
            log_stats,
            executor_fail_callback,
        )

        engine_cores.append(engine_core)
        logger.warning("Disaggregated engine core created.")

    return engine_cores


def _get_slice_sizes(devices):
    prefill_slice_sizes = disagg_utils.get_prefill_slices()
    decode_slice_sizes = disagg_utils.get_decode_slices()
    if isinstance(prefill_slice_sizes[0], int):
        prefill_chip_cnt = sum(prefill_slice_sizes)
    else:
        prefill_chip_cnt = sum([math.prod(t) for t in prefill_slice_sizes])
    if isinstance(decode_slice_sizes[0], int):
        decode_chip_cnt = sum(decode_slice_sizes)
    else:
        decode_chip_cnt = sum([math.prod(t) for t in decode_slice_sizes])
    assert decode_chip_cnt + prefill_chip_cnt <= len(devices)
    assert prefill_chip_cnt > 0 and decode_chip_cnt > 0

    slice_sizes = list(prefill_slice_sizes)
    slice_sizes.extend(decode_slice_sizes)
    return prefill_slice_sizes, decode_slice_sizes, slice_sizes


class DisaggEngineCore(vLLMEngineCore):
    """The vLLM-facing adapter that handles process management and I/O. Modifes vLLMEngineCore and is only used in in-process EngineCore client."""

    @staticmethod
    def is_supported() -> bool:
        """
        Returns True if this engine can run in the current environment.
        """
        return disagg_utils.is_disagg_enabled()

    def __init__(
        self,
        vllm_config: VllmConfig,
        executor_class: type[Executor],
        log_stats: bool,
        executor_fail_callback: Optional[Callable] = None,
    ):
        self.vllm_config = vllm_config

        self.output_queue = queue.Queue[Union[tuple[int, EngineCoreOutputs],
                                              bytes]]()

        self.devices = jax.devices()
        device_kind = self.devices[0].device_kind
        if device_kind != 'TPU7x':
            self.vllm_config.cache_config.gpu_memory_utilization = (
                self.vllm_config.cache_config.gpu_memory_utilization - 0.1)
        prefill_slice_sizes, decode_slice_sizes, slice_sizes = _get_slice_sizes(
            self.devices)

        if isinstance(slice_sizes[0], int):
            setattr(vllm_config.device_config, "slice",
                    (0, slice_sizes, self.devices))
        else:
            setattr(vllm_config.device_config, "slice",
                    ((0, 0), 0, slice_sizes, self.devices))
        logger.info(
            f"Creating DisaggEngineCore with slice_sizes {slice_sizes}...")

        self._prefill_engines = _create_engine_cores(
            prefill_slice_sizes,
            vllm_config,
            log_stats,
            executor_fail_callback,
        )
        logger.info(
            f"{len(self._prefill_engines)} Disaggregated prefill engines created."
        )

        self._decode_engines = _create_engine_cores(
            decode_slice_sizes,
            vllm_config,
            log_stats,
            executor_fail_callback,
        )
        logger.info(
            f"{len(self._decode_engines)} Disaggregated decode engines created."
        )

        self.batch_queue = None

        self.request_block_hasher = None
        if (self.vllm_config.cache_config.enable_prefix_caching
                or self._prefill_engines[0].scheduler.get_kv_connector()
                is not None):

            block_size = vllm_config.cache_config.block_size
            caching_hash_fn = common_utils.get_hash_fn_by_name(
                vllm_config.cache_config.prefix_caching_hash_algo)
            init_none_hash(caching_hash_fn)

            self.request_block_hasher = get_request_block_hasher(
                block_size, caching_hash_fn)

        self.step_fn = (self.step if self.batch_queue is None else
                        self.step_with_batch_queue)

        self.mm_receiver_cache = None
        self._orchestrator = _DisaggOrchestrator(
            config=vllm_config,
            output_queue=self.output_queue,
            prefill_engines=self._prefill_engines,
            decode_engines=self._decode_engines,
            prefill_slice_sizes=prefill_slice_sizes,
            decode_slice_sizes=decode_slice_sizes,
        )
        # for vllm compatibility
        self.model_executor = self._prefill_engines[0].model_executor

    def get_supported_tasks(self) -> tuple[SupportedTask, ...]:
        return self._prefill_engines[0].model_executor.supported_tasks

    def add_request(self, request: Request, request_wave: int = 0):
        if not isinstance(request.request_id, str):
            raise TypeError(
                f"request_id must be a string, got {type(request.request_id)}")

        if pooling_params := request.pooling_params:
            supported_pooling_tasks = [
                task for task in self.get_supported_tasks()
                if task in POOLING_TASKS
            ]

            if pooling_params.task not in supported_pooling_tasks:
                raise ValueError(f"Unsupported task: {pooling_params.task!r} "
                                 f"Supported tasks: {supported_pooling_tasks}")

        if request.kv_transfer_params is not None and (
                not self.scheduler.get_kv_connector()):
            logger.warning("Got kv_transfer_params, but no KVConnector found. "
                           "Disabling KVTransfer for this request.")

        self._orchestrator.add_request(request)

    def step(self) -> tuple[dict[int, EngineCoreOutputs], bool]:
        client_idx, output = self.output_queue.get()
        # logger.warning(f"step output: {output}")
        time.sleep(0.03)
        return {client_idx: output}, True

    def shutdown(self):
        self._orchestrator.shutdown()

    def reset_mm_cache(self):
        # NOTE: Since this is mainly for debugging, we don't attempt to
        # re-sync the internal caches (P0 processor, P0 mirror, P1 mirror)
        for engine in itertools.chain(self._prefill_engines,
                                      self._decode_engines):
            if engine.scheduler.has_unfinished_requests():
                logger.warning(
                    "Resetting the multi-modal cache when requests are "
                    "in progress may lead to desynced internal caches.")

            if engine.mm_receiver_cache is not None:
                engine.mm_receiver_cache.clear_cache()

    def reset_prefix_cache(self):
        for engine in itertools.chain(self._prefill_engines,
                                      self._decode_engines):
            engine.scheduler.reset_prefix_cache()

    def delete_kv_cache(self) -> None:
        for engine in itertools.chain(self._prefill_engines,
                                      self._decode_engines):
            engine.collective_rpc("delete_kv_cache")

    def reinitialize_kv_cache(self) -> None:
        for engine in itertools.chain(self._prefill_engines,
                                      self._decode_engines):
            engine.collective_rpc("reinitialize_kv_cache")


class DisaggEngineCoreProc(vLLMEngineCoreProc):
    """The vLLM-facing adapter that handles process management and I/O."""

    @staticmethod
    def is_supported() -> bool:
        """
        Returns True if this engine can run in the current environment.
        """
        return disagg_utils.is_disagg_enabled()

    def __init__(
        self,
        vllm_config: VllmConfig,
        local_client: bool,
        handshake_address: str,
        executor_class: type[Executor],
        log_stats: bool,
        client_handshake_address: Optional[str] = None,
        engine_index: int = 0,
        **kwargs,
    ):
        if 'dp_rank' in kwargs or 'local_dp_rank' in kwargs:
            logger.debug(
                "Ignoring data parallelism arguments for non-DP disaggregated engine."
            )
        # We don't invoke super class's ctor as we are not really the
        # engine core to be executed, instead we create other instance of
        # engine cores and let them do the work.
        self.vllm_config = vllm_config

        # We should be taking the input from the client, the code below is forked from
        # vllm.v1.engine.core.EngineCoreProc.
        self.input_queue = queue.Queue[tuple[EngineCoreRequestType, Any]]()
        self.output_queue = queue.Queue[Union[tuple[int, EngineCoreOutputs],
                                              bytes]]()
        self.aborts_queue = queue.Queue()
        self.tensor_ipc_receiver = None

        self.engine_index = engine_index
        identity = self.engine_index.to_bytes(length=2, byteorder="little")
        self.engines_running = False

        self.devices = jax.devices()
        device_kind = self.devices[0].device_kind
        if device_kind != 'TPU7x':
            self.vllm_config.cache_config.gpu_memory_utilization = (
                self.vllm_config.cache_config.gpu_memory_utilization - 0.1)
        prefill_slice_sizes, decode_slice_sizes, slice_sizes = _get_slice_sizes(
            self.devices)

        if isinstance(slice_sizes[0], int):
            setattr(vllm_config.device_config, "slice",
                    (0, slice_sizes, self.devices))
        else:
            setattr(vllm_config.device_config, "slice",
                    ((0, 0), 0, slice_sizes, self.devices))
        logger.info(
            f"Creating DisaggEngineCoreProc with slice_sizes {slice_sizes}...")

        def executor_fail_callback():
            self.input_queue.put_nowait(
                (EngineCoreRequestType.EXECUTOR_FAILED, b''))

        # Don't complete handshake until DP coordinator ready message is
        # received.
        with self._perform_handshakes(handshake_address, identity,
                                      local_client, vllm_config,
                                      client_handshake_address) as addresses:
            self.client_count = len(addresses.outputs)

            # Set up data parallel environment.
            self.has_coordinator = addresses.coordinator_output is not None
            self.frontend_stats_publish_address = (
                addresses.frontend_stats_publish_address)
            self.publish_dp_lb_stats = (
                self.has_coordinator
                and not vllm_config.parallel_config.data_parallel_external_lb)
            # Background Threads and Queues for IO. These enable us to
            # overlap ZMQ socket IO with GPU since they release the GIL,
            # and to overlap some serialization/deserialization with the
            # model forward pass.
            # Threads handle Socket <-> Queues and core_busy_loop uses Queue.

            self._prefill_engines = _create_engine_cores(
                prefill_slice_sizes,
                vllm_config,
                log_stats,
                executor_fail_callback,
            )
            logger.info(
                f"{len(self._prefill_engines)} Disaggregated prefill engines created."
            )

            self._decode_engines = _create_engine_cores(
                decode_slice_sizes,
                vllm_config,
                log_stats,
                executor_fail_callback,
            )
            logger.info(
                f"{len(self._decode_engines)} Disaggregated decode engines created."
            )

            try:
                from tpu_inference.runner.continuous_block_pool import \
                    ContinuousFreeQueue
                for core in self._prefill_engines + self._decode_engines:
                    pool = core.scheduler.kv_cache_manager.block_pool
                    pool.free_block_queue = ContinuousFreeQueue(pool.blocks)
                    pool.null_block = pool.free_block_queue.popleft()
                    pool.null_block.is_null = True
                logger.warning(
                    "Successfully injected ContinuousFreeQueue into DisaggEngineCoreProcs."
                )
            except Exception as e:
                logger.warning(f"Failed to inject ContinuousFreeQueue: {e}")

            ready_event = threading.Event()
            input_thread = threading.Thread(target=self.process_input_sockets,
                                            args=(addresses.inputs,
                                                  addresses.coordinator_input,
                                                  identity, ready_event),
                                            daemon=True)
            input_thread.start()

            self.output_thread = threading.Thread(
                target=self.process_output_sockets,
                args=(addresses.outputs, addresses.coordinator_output,
                      self.engine_index),
                daemon=True)
            self.output_thread.start()
            while not ready_event.wait(timeout=10):
                if not input_thread.is_alive():
                    raise RuntimeError(
                        "Input socket thread died during startup")
                if addresses.coordinator_input is not None:
                    logger.info(
                        "Waiting for READY message from DP Coordinator...")
        self.request_block_hasher = None
        if (self.vllm_config.cache_config.enable_prefix_caching
                or self._prefill_engines[0].scheduler.get_kv_connector()
                is not None):

            block_size = vllm_config.cache_config.block_size
            caching_hash_fn = common_utils.get_hash_fn_by_name(
                vllm_config.cache_config.prefix_caching_hash_algo)
            init_none_hash(caching_hash_fn)

            self.request_block_hasher = get_request_block_hasher(
                block_size, caching_hash_fn)

        self.mm_receiver_cache = None
        self._orchestrator = _DisaggOrchestrator(
            config=vllm_config,
            output_queue=self.output_queue,
            prefill_engines=self._prefill_engines,
            decode_engines=self._decode_engines,
            prefill_slice_sizes=prefill_slice_sizes,
            decode_slice_sizes=decode_slice_sizes,
        )

    def add_request(self, request: EngineCoreRequest, request_wave: int = 0):
        if not isinstance(request.request_id, str):
            raise TypeError(
                f"request_id must be a string, got {type(request.request_id)}")

        if pooling_params := request.pooling_params:
            supported_pooling_tasks = [
                task for task in self.get_supported_tasks()
                if task in POOLING_TASKS
            ]

            if pooling_params.task not in supported_pooling_tasks:
                raise ValueError(f"Unsupported task: {pooling_params.task!r} "
                                 f"Supported tasks: {supported_pooling_tasks}")

        self._orchestrator.add_request(request)

    def _handle_client_request(self, request_type: EngineCoreRequestType,
                               request: Any) -> None:
        """Dispatch request from client."""
        if request_type == EngineCoreRequestType.ADD:
            req, request_wave = request
            self.add_request(req)
        elif request_type == EngineCoreRequestType.ABORT:
            # TODO(fhzhang): we need to keep track of which engine is processing
            # the request and finish it there.
            # owner_engine.scheduler.finish_requests(request, RequestStatus.FINISHED_ABORTED)
            pass
        elif request_type == EngineCoreRequestType.UTILITY:
            client_idx, call_id, method_name, args = request
            output = UtilityOutput(call_id)
            try:
                method = getattr(self._prefill_engines[0], method_name)
                result = method(*self._convert_msgspec_args(method, args))
                output.result = UtilityResult(result)
            except BaseException as e:
                logger.exception("Invocation of %s method failed", method_name)
                output.failure_message = (f"Call to {method_name} method"
                                          f" failed: {str(e)}")
            self.output_queue.put_nowait(
                (client_idx, EngineCoreOutputs(utility_output=output)))
        elif request_type == EngineCoreRequestType.EXECUTOR_FAILED:
            raise RuntimeError("Executor failed.")
        else:
            logger.error("Unrecognized input request type encountered: %s",
                         request_type)

    def run_busy_loop(self):
        """Core busy loop of the EngineCore."""

        # Loop until process is sent a SIGINT or SIGTERM
        while True:
            while not self.input_queue.empty():
                req = self.input_queue.get_nowait()
                self._handle_client_request(*req)
            # Yield control to other threads, as we are not doing any real work.
            # Without this sleep, we'd be hogging all the cpu cycles with our run_busy_loop.
            time.sleep(0.01)

    def shutdown(self):
        self._orchestrator.shutdown()

    def delete_kv_cache(self) -> None:
        """Delete KV cache on all workers."""
        for engine in itertools.chain(self._prefill_engines,
                                      self._decode_engines):
            engine.collective_rpc("delete_kv_cache")

    def reinitialize_kv_cache(self) -> None:
        """Reallocate KV cache on all workers."""
        for engine in itertools.chain(self._prefill_engines,
                                      self._decode_engines):
            engine.collective_rpc("reinitialize_kv_cache")



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Type

from jax.sharding import Mesh
from vllm.config import VllmConfig
from vllm.model_executor.layers.quantization.base_config import \
    QuantizationConfig

from tpu_inference.layers.common import quant_methods
from tpu_inference.layers.vllm.quantization.awq import VllmAWQConfig
from tpu_inference.layers.vllm.quantization.compressed_tensors.compressed_tensors import \
    VllmCompressedTensorsConfig
from tpu_inference.layers.vllm.quantization.configs import VllmQuantConfig
from tpu_inference.layers.vllm.quantization.fp8 import VllmFp8Config
from tpu_inference.layers.vllm.quantization.mxfp4 import VllmMxfp4Config
from tpu_inference.layers.vllm.quantization.unquantized import \
    VllmUnquantizedConfig


def get_tpu_quantization_config(vllm_config: VllmConfig,
                                mesh: Mesh) -> QuantizationConfig:
    model_config = copy.deepcopy(vllm_config.model_config)
    # TODO(kyuyeunk): Add support for "tpu_int8".
    method_to_config: dict[str | None, Type[QuantizationConfig]] = {
        None: VllmUnquantizedConfig,
        quant_methods.COMPRESSED_TENSORS: VllmCompressedTensorsConfig,
        quant_methods.AWQ: VllmAWQConfig,
        quant_methods.FP8: VllmFp8Config,
        quant_methods.MXFP4: VllmMxfp4Config,
    }
    if model_config.quantization not in method_to_config:
        raise NotImplementedError(
            f"{model_config.quantization} quantization method not supported."
            f" Supported methods are {method_to_config.keys()}")
    quant_config = method_to_config[model_config.quantization]
    assert issubclass(quant_config, VllmQuantConfig)
    quant_config.set_configs(vllm_config, mesh)

    model_config.quantization = quant_config.get_name()
    return VllmConfig.get_quantization_config(model_config,
                                              vllm_config.load_config)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for pathways_dummy_loader module."""

from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from flax import nnx
from jax.sharding import (Mesh, NamedSharding, PartitionSpec,
                          SingleDeviceSharding)

from tpu_inference.layers.jax import JaxModule, JaxModuleList
from tpu_inference.models.common.pathways_dummy_loader import (
    _HIGH, _LOW, PathwaysDummyModelLoader, _process_weights_after_loading_jax,
    create_dummy_weights_on_tpu, is_pathways_dummy_load,
    load_dummy_weights_jax)

# ==============================================================================
# >> Fixtures
# ==============================================================================


@pytest.fixture(scope="module")
def mesh() -> Mesh:
    """Provides a single-device JAX mesh for testing."""
    devices = np.array(jax.devices()[:1]).reshape((1, 1))
    return Mesh(devices, axis_names=("data", "model"))


@pytest.fixture
def single_device_mesh() -> Mesh:
    """Provides a 1-device mesh with standard axis names."""
    devices = np.array(jax.devices()[:1]).reshape((1, ))
    return Mesh(devices, axis_names=("model", ))


# ==============================================================================
# >> Tests for is_pathways_dummy_load
# ==============================================================================


class TestIsPathwaysDummyLoad:
    """Tests for the is_pathways_dummy_load() helper function."""

    @patch("tpu_inference.models.common.pathways_dummy_loader.vllm_envs")
    def test_returns_false_when_not_using_pathways(self, mock_envs):
        """Should return False when VLLM_TPU_USING_PATHWAYS is False."""
        mock_envs.VLLM_TPU_USING_PATHWAYS = False
        assert is_pathways_dummy_load() is False

    @patch(
        "tpu_inference.models.common.pathways_dummy_loader.get_current_vllm_config"
    )
    @patch("tpu_inference.models.common.pathways_dummy_loader.vllm_envs")
    def test_returns_true_for_dummy_format(self, mock_envs, mock_get_config):
        """Should return True when using Pathways with load_format='dummy'."""
        mock_envs.VLLM_TPU_USING_PATHWAYS = True
        mock_config = MagicMock()
        mock_config.load_config.load_format = "dummy"
        mock_get_config.return_value = mock_config
        assert is_pathways_dummy_load() is True

    @patch(
        "tpu_inference.models.common.pathways_dummy_loader.get_current_vllm_config"
    )
    @patch("tpu_inference.models.common.pathways_dummy_loader.vllm_envs")
    def test_returns_true_for_pathways_dummy_format(self, mock_envs,
                                                    mock_get_config):
        """Should return True when using Pathways with load_format='pathways_dummy'."""
        mock_envs.VLLM_TPU_USING_PATHWAYS = True
        mock_config = MagicMock()
        mock_config.load_config.load_format = "pathways_dummy"
        mock_get_config.return_value = mock_config
        assert is_pathways_dummy_load() is True

    @patch(
        "tpu_inference.models.common.pathways_dummy_loader.get_current_vllm_config"
    )
    @patch("tpu_inference.models.common.pathways_dummy_loader.vllm_envs")
    def test_returns_false_for_other_load_formats(self, mock_envs,
                                                  mock_get_config):
        """Should return False for non-dummy load formats even with Pathways."""
        mock_envs.VLLM_TPU_USING_PATHWAYS = True
        mock_config = MagicMock()
        mock_config.load_config.load_format = "auto"
        mock_get_config.return_value = mock_config
        assert is_pathways_dummy_load() is False


# ==============================================================================
# >> Tests for create_dummy_weights_on_tpu
# ==============================================================================


class TestCreateDummyWeightsOnTpu:
    """Tests for the create_dummy_weights_on_tpu() function."""

    def test_returns_correct_shape(self, mesh):
        """Generated array should have the requested shape."""
        shape = (4, 8)
        sharding = NamedSharding(mesh, PartitionSpec())
        result = create_dummy_weights_on_tpu(sharding=sharding,
                                             weight_shape=shape,
                                             weight_dtype=jnp.float32)
        assert result.shape == shape

    def test_returns_correct_dtype(self, mesh):
        """Generated array should have the requested dtype."""
        sharding = NamedSharding(mesh, PartitionSpec())
        for dtype in [jnp.float32, jnp.bfloat16, jnp.float16]:
            result = create_dummy_weights_on_tpu(sharding=sharding,
                                                 weight_shape=(2, 3),
                                                 weight_dtype=dtype)
            assert result.dtype == dtype

    def test_values_within_range(self, mesh):
        """All values should be in [_LOW, _HIGH]."""
        sharding = NamedSharding(mesh, PartitionSpec())
        result = create_dummy_weights_on_tpu(sharding=sharding,
                                             weight_shape=(100, 100),
                                             weight_dtype=jnp.float32)
        values = np.array(result)
        assert np.all(values >= _LOW)
        assert np.all(values <= _HIGH)

    def test_deterministic_with_same_seed(self, mesh):
        """Calling twice should produce identical results (fixed seed)."""
        sharding = NamedSharding(mesh, PartitionSpec())
        a = create_dummy_weights_on_tpu(sharding=sharding,
                                        weight_shape=(4, 4),
                                        weight_dtype=jnp.float32)
        b = create_dummy_weights_on_tpu(sharding=sharding,
                                        weight_shape=(4, 4),
                                        weight_dtype=jnp.float32)
        np.testing.assert_array_equal(np.array(a), np.array(b))

    def test_sharding_is_applied(self, mesh):
        """Result should carry the requested sharding."""
        spec = PartitionSpec("data", "model")
        sharding = NamedSharding(mesh, spec)
        result = create_dummy_weights_on_tpu(sharding=sharding,
                                             weight_shape=(4, 8),
                                             weight_dtype=jnp.float32)
        assert result.sharding == sharding

    def test_scalar_shape(self, mesh):
        """Should handle scalar (empty-tuple) shapes."""
        sharding = NamedSharding(mesh, PartitionSpec())
        result = create_dummy_weights_on_tpu(sharding=sharding,
                                             weight_shape=(),
                                             weight_dtype=jnp.float32)
        assert result.shape == ()

    def test_1d_shape(self, single_device_mesh):
        """Should handle 1-D shapes."""
        sharding = NamedSharding(single_device_mesh, PartitionSpec("model"))
        result = create_dummy_weights_on_tpu(sharding=sharding,
                                             weight_shape=(16, ),
                                             weight_dtype=jnp.float32)
        assert result.shape == (16, )


# ==============================================================================
# >> Tests for load_dummy_weights_jax
# ==============================================================================


class _SimpleLinear(JaxModule):
    """A minimal JaxModule with a single weight param for testing."""

    def __init__(self, in_features: int, out_features: int, dtype=jnp.float32):
        super().__init__()
        abstract = jnp.zeros((in_features, out_features), dtype=dtype)
        self.weight = nnx.Param(abstract)
        self.weight.set_metadata("sharding", PartitionSpec())

    def __call__(self, x):
        return x @ self.weight.value


class _NestedModel(JaxModule):
    """A JaxModule containing a JaxModuleList of _SimpleLinear."""

    def __init__(self, dtype=jnp.float32):
        super().__init__()
        self.layers = JaxModuleList([
            _SimpleLinear(4, 8, dtype=dtype),
            _SimpleLinear(8, 4, dtype=dtype),
        ])

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x


class TestLoadDummyWeightsJax:
    """Tests for load_dummy_weights_jax()."""

    def test_fills_simple_model_weights(self, mesh):
        """All params should be filled with non-zero dummy data."""
        model = _SimpleLinear(4, 8)
        # Initially all zeros
        assert np.all(np.array(model.weight.value) == 0.0)

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader._process_weights_after_loading_jax"
        ):
            load_dummy_weights_jax(model, mesh)

        # After loading, should not be all zeros
        values = np.array(model.weight.value)
        assert values.shape == (4, 8)
        assert np.all(values >= _LOW)
        assert np.all(values <= _HIGH)

    def test_fills_nested_model_weights(self, mesh):
        """Nested model params should all be filled."""
        model = _NestedModel()

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader._process_weights_after_loading_jax"
        ):
            load_dummy_weights_jax(model, mesh)

        for _name, param in model.named_parameters():
            values = np.array(param.value)
            assert np.all(values >= _LOW)
            assert np.all(values <= _HIGH)

    def test_calls_process_weights_after_loading(self, mesh):
        """Should call _process_weights_after_loading_jax after loading."""
        model = _SimpleLinear(4, 8)

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader._process_weights_after_loading_jax"
        ) as mock_process:
            load_dummy_weights_jax(model, mesh)
            mock_process.assert_called_once_with(model)

    def test_respects_param_dtype(self, mesh):
        """Loaded weights should have the same dtype as the original param."""
        model = _SimpleLinear(4, 8, dtype=jnp.bfloat16)

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader._process_weights_after_loading_jax"
        ):
            load_dummy_weights_jax(model, mesh)

        assert model.weight.value.dtype == jnp.bfloat16

    def test_handles_named_sharding_spec(self, mesh):
        """Params with NamedSharding metadata should be handled correctly."""
        model = _SimpleLinear(4, 8)
        sharding = NamedSharding(mesh, PartitionSpec("data", "model"))
        model.weight.set_metadata("sharding", sharding)

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader._process_weights_after_loading_jax"
        ):
            load_dummy_weights_jax(model, mesh)

        values = np.array(model.weight.value)
        assert values.shape == (4, 8)

    def test_handles_single_device_sharding_spec(self, mesh):
        """Params with SingleDeviceSharding metadata should be handled correctly."""
        model = _SimpleLinear(4, 8)
        device = jax.devices()[0]
        model.weight.set_metadata("sharding", SingleDeviceSharding(device))

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader._process_weights_after_loading_jax"
        ):
            load_dummy_weights_jax(model, mesh)

        values = np.array(model.weight.value)
        assert values.shape == (4, 8)


# ==============================================================================
# >> Tests for _process_weights_after_loading_jax
# ==============================================================================


class TestProcessWeightsAfterLoadingJax:
    """Tests for _process_weights_after_loading_jax()."""

    def test_calls_quant_method_process_weights(self):
        """Should call process_weights_after_loading on quant method."""
        from tpu_inference.layers.jax.quantization import QuantizeMethodBase

        # Create a concrete subclass of QuantizeMethodBase so isinstance checks pass
        class FakeQuantMethod(QuantizeMethodBase):

            def create_weights(self, *args, **kwargs):
                pass

            def apply(self, *args, **kwargs):
                pass

            def process_weights_after_loading(self, module):
                pass

        mock_quant = MagicMock(spec=FakeQuantMethod)
        # Make isinstance(mock_quant, QuantizeMethodBase) return True
        mock_quant.__class__ = FakeQuantMethod

        module = MagicMock(spec=JaxModule)
        module.quant_method = mock_quant

        _process_weights_after_loading_jax(module)

        mock_quant.process_weights_after_loading.assert_called_once_with(
            module)

    def test_recurses_through_jax_module_list(self):
        """Should recursively process children in JaxModuleList."""
        child1 = MagicMock(spec=JaxModule)
        child1.quant_method = None
        child1.named_children.return_value = []

        child2 = MagicMock(spec=JaxModule)
        child2.quant_method = None
        child2.named_children.return_value = []

        module_list = JaxModuleList([child1, child2])

        # Should not raise
        _process_weights_after_loading_jax(module_list)

    def test_recurses_through_named_children(self):
        """Should recursively process named_children on non-list modules."""
        child = MagicMock(spec=JaxModule)
        child.quant_method = None
        child.named_children.return_value = []

        parent = MagicMock(spec=JaxModule)
        parent.quant_method = None
        parent.named_children.return_value = [("child", child)]

        # Should not raise
        _process_weights_after_loading_jax(parent)


# ==============================================================================
# >> Tests for PathwaysDummyModelLoader
# ==============================================================================


class TestPathwaysDummyModelLoader:
    """Tests for the PathwaysDummyModelLoader class."""

    def test_download_model_is_noop(self):
        """download_model should do nothing."""
        load_config = MagicMock()
        loader = PathwaysDummyModelLoader(load_config)
        # Should not raise
        loader.download_model(model_config=MagicMock())

    def test_load_weights_dispatches_to_jax_backend(self, mesh):
        """load_weights should call load_dummy_weights_jax for JaxModule models."""
        load_config = MagicMock()
        loader = PathwaysDummyModelLoader(load_config)
        model = _SimpleLinear(4, 8)

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader.load_dummy_weights_jax"
        ) as mock_load_jax:
            with patch(
                    "tpu_inference.models.common.pathways_dummy_loader.jax.sharding.get_mesh",
                    return_value=mesh,
            ):
                loader.load_weights(model, model_config=MagicMock())

            mock_load_jax.assert_called_once_with(model, mesh)

            mock_load_jax.assert_called_once_with(model, mesh)

    def test_load_weights_noop_for_non_jax_model(self):
        """load_weights should be a no-op for non-JaxModule (torchax) models."""
        load_config = MagicMock()
        loader = PathwaysDummyModelLoader(load_config)
        model = MagicMock()  # Not a JaxModule

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader.load_dummy_weights_jax"
        ) as mock_load_jax:
            loader.load_weights(model, model_config=MagicMock())
            mock_load_jax.assert_not_called()

    def test_load_weights_raises_without_mesh_for_jax(self):
        """load_weights should raise RuntimeError when no mesh is set for JaxModule."""
        load_config = MagicMock()
        loader = PathwaysDummyModelLoader(load_config)
        model = _SimpleLinear(4, 8)

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader.jax.sharding.get_mesh",
                return_value=None,
        ):
            with pytest.raises(RuntimeError, match="active JAX mesh"):
                loader.load_weights(model, model_config=MagicMock())

    def test_load_model_initializes_and_processes_weights(self):
        """load_model should initialize model and call process_weights_after_loading."""
        load_config = MagicMock()
        load_config.device = None
        loader = PathwaysDummyModelLoader(load_config)

        mock_vllm_config = MagicMock()
        mock_vllm_config.device_config.device = "cpu"
        mock_vllm_config.load_config = load_config

        mock_model_config = MagicMock()
        mock_model_config.dtype = torch.float32

        mock_model = MagicMock()
        mock_model.eval.return_value = mock_model

        with patch(
                "tpu_inference.models.common.pathways_dummy_loader.initialize_model",
                return_value=mock_model,
        ) as mock_init:
            with patch(
                    "tpu_inference.models.common.pathways_dummy_loader.process_weights_after_loading"
            ) as mock_process:
                result = loader.load_model(
                    vllm_config=mock_vllm_config,
                    model_config=mock_model_config,
                )

        mock_init.assert_called_once()
        mock_process.assert_called_once()
        mock_model.eval.assert_called_once()
        assert result is mock_model


# ==============================================================================
# >> Registration test
# ==============================================================================


class TestRegistration:
    """Tests that the loader is properly registered."""

    def test_pathways_dummy_loader_is_registered(self):
        """PathwaysDummyModelLoader should be discoverable via the registry."""
        from vllm.model_executor.model_loader import get_model_loader

        load_config = MagicMock()
        load_config.load_format = "pathways_dummy"
        loader = get_model_loader(load_config)
        assert isinstance(loader, PathwaysDummyModelLoader)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import pytest
from vllm.config import VllmConfig
from vllm.v1.core.sched.interface import PauseState
from vllm.v1.core.sched.output import CachedRequestData, SchedulerOutput
from vllm.v1.core.sched.scheduler import Scheduler
from vllm.v1.kv_cache_interface import KVCacheConfig
from vllm.v1.metrics.stats import PrefixCacheStats, SchedulerStats
from vllm.v1.request import Request

from tpu_inference.core.sched.dp_scheduler import (
    DPScheduler, DPSchedulerOutput, SchedulerCommand,
    update_vllm_config_for_dp_scheduler)


def _make_mock_mp_context():
    """Create a mock multiprocessing context that properly handles Pipe()."""
    mock_ctx = MagicMock()
    mock_process = MagicMock()
    mock_ctx.Pipe = MagicMock(return_value=(MagicMock(), MagicMock()))
    mock_ctx.Process = MagicMock(return_value=mock_process)
    return mock_ctx


class TestDPScheduler:

    @pytest.fixture
    def mock_vllm_config(self):
        """Create a mock VllmConfig for testing."""
        config = MagicMock(spec=VllmConfig)
        config.sharding_config = MagicMock()
        config.sharding_config.total_dp_size = 2
        config.scheduler_config = MagicMock()
        config.scheduler_config._original_scheduler_cls = Scheduler
        config.scheduler_config.max_num_seqs = 8
        config.scheduler_config.max_num_batched_tokens = 1024
        config.scheduler_config.async_scheduling = False
        config.cache_config = MagicMock()
        config.cache_config.enable_prefix_caching = False
        config.cache_config.prefix_caching_hash_algo = "sha256"
        return config

    @pytest.fixture
    def mock_kv_cache_config(self):
        """Create a mock KVCacheConfig for testing."""
        config = MagicMock(spec=KVCacheConfig)
        config.num_blocks = 100
        return config

    @pytest.fixture
    def mock_structured_output_manager(self):
        """Create a mock StructuredOutputManager."""
        return MagicMock()

    def _create_scheduler(self,
                          mock_vllm_config,
                          mock_kv_cache_config,
                          mock_structured_output_manager,
                          log_stats=False):
        """Helper to create a DPScheduler with mocked multiprocessing."""
        mock_ctx = _make_mock_mp_context()
        with patch(
                'tpu_inference.core.sched.dp_scheduler._scheduler_worker_process'
        ):
            with patch('multiprocessing.get_context', return_value=mock_ctx):
                scheduler = DPScheduler(
                    vllm_config=mock_vllm_config,
                    kv_cache_config=mock_kv_cache_config,
                    structured_output_manager=mock_structured_output_manager,
                    block_size=16,
                    log_stats=log_stats,
                )
        return scheduler

    def test_init_creates_worker_processes(
        self,
        mock_vllm_config,
        mock_kv_cache_config,
        mock_structured_output_manager,
    ):
        """Test initialization creates worker processes for each DP rank."""
        with patch(
                'tpu_inference.core.sched.dp_scheduler._scheduler_worker_process'
        ):
            with patch('multiprocessing.get_context') as mock_get_context:
                mock_ctx = _make_mock_mp_context()
                mock_get_context.return_value = mock_ctx

                scheduler = DPScheduler(
                    vllm_config=mock_vllm_config,
                    kv_cache_config=mock_kv_cache_config,
                    structured_output_manager=mock_structured_output_manager,
                    block_size=16,
                    log_stats=True,
                )

                # Verify processes and connections were created
                assert scheduler.dp_size == 2
                assert len(scheduler.processes) == 2
                # One input conn and one output conn per rank
                assert len(scheduler.input_conns) == 2
                assert len(scheduler.output_conns) == 2
                assert scheduler.log_stats is True
                assert len(scheduler.per_rank_kv_cache_configs) == 2

                # Verify each rank got the correct config
                for rank_config in scheduler.per_rank_kv_cache_configs:
                    assert rank_config.num_blocks == 50  # 100 / 2

                # Verify processes were started
                mock_process = mock_ctx.Process.return_value
                assert mock_process.start.call_count == 2

    def test_init_with_prefix_caching_enabled(
        self,
        mock_vllm_config,
        mock_kv_cache_config,
        mock_structured_output_manager,
    ):
        """Test initialization with prefix caching enabled initializes NONE_HASH."""
        mock_vllm_config.cache_config.enable_prefix_caching = True
        mock_vllm_config.cache_config.prefix_caching_hash_algo = "sha256"

        with patch(
                'tpu_inference.core.sched.dp_scheduler._scheduler_worker_process'
        ):
            with patch('multiprocessing.get_context') as mock_get_context:
                with patch('vllm.v1.core.kv_cache_utils.init_none_hash'
                           ) as mock_init_none_hash:
                    with patch('vllm.utils.hashing.get_hash_fn_by_name'
                               ) as mock_get_hash_fn:
                        mock_ctx = _make_mock_mp_context()
                        mock_get_context.return_value = mock_ctx

                        mock_hash_fn = MagicMock()
                        mock_get_hash_fn.return_value = mock_hash_fn

                        scheduler = DPScheduler(
                            vllm_config=mock_vllm_config,
                            kv_cache_config=mock_kv_cache_config,
                            structured_output_manager=
                            mock_structured_output_manager,
                            block_size=16,
                            log_stats=True,
                        )

                        # Verify init_none_hash was called with correct hash function
                        mock_get_hash_fn.assert_called_once_with("sha256")
                        mock_init_none_hash.assert_called_once_with(
                            mock_hash_fn)

                        assert scheduler.dp_size == 2

    def test_init_without_prefix_caching_skips_initialization(
        self,
        mock_vllm_config,
        mock_kv_cache_config,
        mock_structured_output_manager,
    ):
        """Test initialization without prefix caching skips NONE_HASH initialization."""
        mock_vllm_config.cache_config.enable_prefix_caching = False

        with patch(
                'tpu_inference.core.sched.dp_scheduler._scheduler_worker_process'
        ):
            with patch('multiprocessing.get_context') as mock_get_context:
                with patch('vllm.v1.core.kv_cache_utils.init_none_hash'
                           ) as mock_init_none_hash:
                    mock_ctx = _make_mock_mp_context()
                    mock_get_context.return_value = mock_ctx

                    scheduler = DPScheduler(
                        vllm_config=mock_vllm_config,
                        kv_cache_config=mock_kv_cache_config,
                        structured_output_manager=
                        mock_structured_output_manager,
                        block_size=16,
                        log_stats=True,
                    )

                    # Verify init_none_hash was NOT called
                    mock_init_none_hash.assert_not_called()

                    assert scheduler.dp_size == 2

    def test_get_rank_token_counts(self, mock_vllm_config,
                                   mock_kv_cache_config,
                                   mock_structured_output_manager):
        """Test _get_rank_token_counts queries workers and aggregates tokens."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        # Mock _send_command and _get_result (the pipe-based API)
        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[30, 15])

        rank_tokens = scheduler._get_rank_token_counts()

        # Verify correct commands were sent
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.GET_TOKEN_COUNT)
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.GET_TOKEN_COUNT)

        # Verify results were collected
        scheduler._get_result.assert_any_call(0,
                                              SchedulerCommand.GET_TOKEN_COUNT)
        scheduler._get_result.assert_any_call(1,
                                              SchedulerCommand.GET_TOKEN_COUNT)

        assert rank_tokens[0] == 30
        assert rank_tokens[1] == 15

    def test_find_best_rank_with_cache_hit(self, mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager):
        """Test _find_best_rank_for_request prefers cache hits."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        mock_request = MagicMock(spec=Request)

        # Mock _send_command and _get_result
        # _get_rank_token_counts calls _get_result for GET_TOKEN_COUNT (2 ranks)
        # then _find_best_rank calls _get_result for PROBE_COMPUTED_BLOCKS (2 ranks)
        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[
            100,
            50,  # GET_TOKEN_COUNT: rank 0=100, rank 1=50
            10,
            25,  # PROBE_COMPUTED_BLOCKS: rank 0=10, rank 1=25
        ])

        rank = scheduler._find_best_rank_for_request(mock_request)

        # Should prefer rank with better cache hit (rank 1 has 25 cached tokens)
        assert rank == 1

    def test_find_best_rank_without_cache_hit(self, mock_vllm_config,
                                              mock_kv_cache_config,
                                              mock_structured_output_manager):
        """Test _find_best_rank_for_request uses load balancing without cache hit."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        mock_request = MagicMock(spec=Request)

        # Mock _send_command and _get_result
        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[
            100,
            50,  # GET_TOKEN_COUNT: rank 0=100, rank 1=50
            0,
            0,  # PROBE_COMPUTED_BLOCKS: no cache hits
        ])

        rank = scheduler._find_best_rank_for_request(mock_request)

        # Should choose rank with fewer tokens (rank 1)
        assert rank == 1

    def test_add_request_assigns_to_best_rank(self, mock_vllm_config,
                                              mock_kv_cache_config,
                                              mock_structured_output_manager):
        """Test add_request assigns request to best rank."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        mock_request = MagicMock(spec=Request)
        mock_request.request_id = "req1"

        # Mock _find_best_rank_for_request to return rank 1
        scheduler._find_best_rank_for_request = MagicMock(return_value=1)
        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=None)

        scheduler.add_request(mock_request)

        # Verify request was assigned to rank 1
        assert scheduler.assigned_dp_rank["req1"] == 1

        # Verify ADD_REQUEST command was sent to rank 1
        scheduler._send_command.assert_called_with(
            1, SchedulerCommand.ADD_REQUEST, mock_request)

        # Verify we waited for completion
        scheduler._get_result.assert_called_with(1,
                                                 SchedulerCommand.ADD_REQUEST)

    def test_schedule_sends_commands_and_combines_output(
            self, mock_vllm_config, mock_kv_cache_config,
            mock_structured_output_manager):
        """Test schedule sends SCHEDULE command to all workers and combines output."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        # Create mock scheduler outputs
        mock_output_0 = MagicMock(spec=SchedulerOutput)
        mock_output_0.scheduled_new_reqs = []
        mock_output_0.num_scheduled_tokens = {"req1": 10}
        mock_output_0.total_num_scheduled_tokens = 10
        mock_output_0.finished_req_ids = set()
        mock_output_0.scheduled_cached_reqs = CachedRequestData(
            req_ids=[],
            resumed_req_ids=[],
            new_token_ids=[],
            all_token_ids={},
            new_block_ids=[],
            num_computed_tokens=[],
            num_output_tokens=[],
        )
        mock_output_0.scheduled_spec_decode_tokens = {}
        mock_output_0.scheduled_encoder_inputs = {}
        mock_output_0.num_common_prefix_blocks = []

        mock_output_1 = MagicMock(spec=SchedulerOutput)
        mock_output_1.scheduled_new_reqs = []
        mock_output_1.num_scheduled_tokens = {"req2": 20}
        mock_output_1.total_num_scheduled_tokens = 20
        mock_output_1.finished_req_ids = set()
        mock_output_1.scheduled_cached_reqs = CachedRequestData(
            req_ids=[],
            resumed_req_ids=[],
            new_token_ids=[],
            all_token_ids={},
            new_block_ids=[],
            num_computed_tokens=[],
            num_output_tokens=[],
        )
        mock_output_1.scheduled_spec_decode_tokens = {}
        mock_output_1.scheduled_encoder_inputs = {}
        mock_output_1.num_common_prefix_blocks = []

        # Mock _send_command and _get_result
        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(
            side_effect=[mock_output_0, mock_output_1])

        # Setup assigned ranks
        scheduler.assigned_dp_rank = {"req1": 0, "req2": 1}

        output = scheduler.schedule()

        # Verify SCHEDULE commands were sent
        scheduler._send_command.assert_any_call(0, SchedulerCommand.SCHEDULE)
        scheduler._send_command.assert_any_call(1, SchedulerCommand.SCHEDULE)

        # Verify combined output
        assert isinstance(output, DPSchedulerOutput)
        assert output.total_num_scheduled_tokens == 30
        assert "req1" in output.num_scheduled_tokens
        assert "req2" in output.num_scheduled_tokens
        assert output.assigned_dp_rank == {"req1": 0, "req2": 1}

    def test_combine_cached_request_data(self, mock_vllm_config,
                                         mock_kv_cache_config,
                                         mock_structured_output_manager):
        """Test _combine_cached_request_data combines data from all ranks."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        # Create mock rank outputs
        output_0 = MagicMock(spec=SchedulerOutput)
        output_0.scheduled_cached_reqs = CachedRequestData(
            req_ids=["req1"],
            resumed_req_ids=["req1"],
            new_token_ids=[[1, 2, 3]],
            all_token_ids={"req1": [1, 2, 3, 4, 5]},
            new_block_ids=[[10, 11]],
            num_computed_tokens=[5],
            num_output_tokens=[3],
        )

        output_1 = MagicMock(spec=SchedulerOutput)
        output_1.scheduled_cached_reqs = CachedRequestData(
            req_ids=["req2"],
            resumed_req_ids=[],
            new_token_ids=[[6, 7]],
            all_token_ids={"req2": [6, 7, 8, 9]},
            new_block_ids=[[20, 21]],
            num_computed_tokens=[4],
            num_output_tokens=[2],
        )

        combined = scheduler._combine_cached_request_data([output_0, output_1])

        # Verify combined data
        assert combined.req_ids == ["req1", "req2"]
        assert combined.resumed_req_ids == ["req1"]
        assert combined.new_token_ids == [[1, 2, 3], [6, 7]]
        assert combined.all_token_ids == {
            "req1": [1, 2, 3, 4, 5],
            "req2": [6, 7, 8, 9]
        }
        assert combined.num_computed_tokens == [5, 4]
        assert combined.num_output_tokens == [3, 2]

    def test_finish_requests_routes_to_workers(self, mock_vllm_config,
                                               mock_kv_cache_config,
                                               mock_structured_output_manager):
        """Test finish_requests sends FINISH_REQUESTS command to appropriate workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=None)

        scheduler.assigned_dp_rank = {"req1": 0, "req2": 1, "req3": 0}

        # Test with list of requests
        scheduler.finish_requests(["req1", "req2"],
                                  finished_status="completed")

        # Verify FINISH_REQUESTS commands were sent to correct ranks
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.FINISH_REQUESTS, (["req1"], "completed"))
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.FINISH_REQUESTS, (["req2"], "completed"))

    def test_get_num_unfinished_requests(self, mock_vllm_config,
                                         mock_kv_cache_config,
                                         mock_structured_output_manager):
        """Test get_num_unfinished_requests queries all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[5, 3])

        total = scheduler.get_num_unfinished_requests()

        # Verify commands were sent
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.GET_NUM_UNFINISHED_REQUESTS)
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.GET_NUM_UNFINISHED_REQUESTS)

        assert total == 8

    def test_has_finished_requests(self, mock_vllm_config,
                                   mock_kv_cache_config,
                                   mock_structured_output_manager):
        """Test has_finished_requests checks all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[False, True])

        result = scheduler.has_finished_requests()

        assert result is True

        # Verify commands were sent
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.HAS_FINISHED_REQUESTS)
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.HAS_FINISHED_REQUESTS)

    def test_get_request_counts(self, mock_vllm_config, mock_kv_cache_config,
                                mock_structured_output_manager):
        """Test get_request_counts queries all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[(2, 1), (1, 3)])

        running, waiting = scheduler.get_request_counts()

        # Verify commands were sent
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.GET_REQUEST_COUNTS)
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.GET_REQUEST_COUNTS)

        assert running == 3
        assert waiting == 4

    def test_reset_prefix_cache(self, mock_vllm_config, mock_kv_cache_config,
                                mock_structured_output_manager):
        """Test reset_prefix_cache sends command to all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[True, True])

        result = scheduler.reset_prefix_cache()

        # Verify commands were sent with default args
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.RESET_PREFIX_CACHE, (False, False))
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.RESET_PREFIX_CACHE, (False, False))

        assert result is True

    def test_reset_prefix_cache_with_args(self, mock_vllm_config,
                                          mock_kv_cache_config,
                                          mock_structured_output_manager):
        """Test reset_prefix_cache forwards reset_running_requests and reset_connector args."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[True, False])

        result = scheduler.reset_prefix_cache(reset_running_requests=True,
                                              reset_connector=True)

        # Verify commands were sent with provided args
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.RESET_PREFIX_CACHE, (True, True))
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.RESET_PREFIX_CACHE, (True, True))

        # One rank returned False, so overall result should be False
        assert result is False

    def test_reset_encoder_cache(self, mock_vllm_config, mock_kv_cache_config,
                                 mock_structured_output_manager):
        """Test reset_encoder_cache sends command to all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=None)

        scheduler.reset_encoder_cache()

        # Verify commands were sent
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.RESET_ENCODER_CACHE)
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.RESET_ENCODER_CACHE)

    def test_pause_state_default(self, mock_vllm_config, mock_kv_cache_config,
                                 mock_structured_output_manager):
        """Test pause_state queries worker and defaults to UNPAUSED."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=PauseState.UNPAUSED)

        assert scheduler.pause_state == PauseState.UNPAUSED

        # Verify GET_PAUSE_STATE was sent only to rank 0
        scheduler._send_command.assert_called_once_with(
            0, SchedulerCommand.GET_PAUSE_STATE)

    def test_set_pause_state(self, mock_vllm_config, mock_kv_cache_config,
                             mock_structured_output_manager):
        """Test set_pause_state sends command to all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=None)

        scheduler.set_pause_state(PauseState.PAUSED_NEW)

        # Verify SET_PAUSE_STATE was sent to all ranks
        scheduler._send_command.assert_any_call(
            0, SchedulerCommand.SET_PAUSE_STATE, PauseState.PAUSED_NEW)
        scheduler._send_command.assert_any_call(
            1, SchedulerCommand.SET_PAUSE_STATE, PauseState.PAUSED_NEW)

    def test_make_stats_aggregates_from_workers(
            self, mock_vllm_config, mock_kv_cache_config,
            mock_structured_output_manager):
        """Test make_stats aggregates statistics from all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager,
                                           log_stats=True)

        # Create mock stats
        stats_0 = SchedulerStats(
            num_running_reqs=3,
            num_waiting_reqs=2,
            kv_cache_usage=0.5,
            prefix_cache_stats=PrefixCacheStats(reset=False,
                                                requests=10,
                                                queries=8,
                                                hits=5),
            connector_prefix_cache_stats=PrefixCacheStats(reset=False,
                                                          requests=5,
                                                          queries=4,
                                                          hits=2),
            spec_decoding_stats=None,
            kv_connector_stats=None,
        )

        stats_1 = SchedulerStats(
            num_running_reqs=4,
            num_waiting_reqs=1,
            kv_cache_usage=0.7,
            prefix_cache_stats=PrefixCacheStats(reset=False,
                                                requests=15,
                                                queries=12,
                                                hits=8),
            connector_prefix_cache_stats=PrefixCacheStats(reset=False,
                                                          requests=6,
                                                          queries=5,
                                                          hits=3),
            spec_decoding_stats=None,
            kv_connector_stats=None,
        )

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(side_effect=[stats_0, stats_1])

        combined_stats = scheduler.make_stats()

        # Verify commands were sent
        scheduler._send_command.assert_any_call(0, SchedulerCommand.MAKE_STATS,
                                                (None, None))
        scheduler._send_command.assert_any_call(1, SchedulerCommand.MAKE_STATS,
                                                (None, None))

        assert combined_stats.num_running_reqs == 7
        assert combined_stats.num_waiting_reqs == 3
        assert combined_stats.kv_cache_usage == 0.6

    def test_make_stats_returns_none_when_disabled(
            self, mock_vllm_config, mock_kv_cache_config,
            mock_structured_output_manager):
        """Test make_stats returns None when logging disabled."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager,
                                           log_stats=False)

        stats = scheduler.make_stats()
        assert stats is None

    def test_update_draft_token_ids(self, mock_vllm_config,
                                    mock_kv_cache_config,
                                    mock_structured_output_manager):
        """Test update_draft_token_ids routes to correct workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=None)

        scheduler.assigned_dp_rank = {"req1": 0, "req2": 1, "req3": 0}

        draft_token_ids = MagicMock()
        draft_token_ids.req_ids = ["req1", "req2", "req3"]
        draft_token_ids.draft_token_ids = [
            [101, 102, 103],
            [201, 202],
            [301, 302, 303, 304],
        ]

        scheduler.update_draft_token_ids(draft_token_ids)

        # Verify commands were sent to both workers (rank 0 and rank 1)
        assert scheduler._send_command.call_count == 2
        # Verify the command type for each call
        for c in scheduler._send_command.call_args_list:
            assert c[0][1] == SchedulerCommand.UPDATE_DRAFT_TOKEN_IDS

    def test_combine_scheduler_outputs_max_tokens(
            self, mock_vllm_config, mock_kv_cache_config,
            mock_structured_output_manager):
        """Test _combine_scheduler_outputs calculates max_num_scheduled_tokens_per_dp_rank."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        # Mock SchedulerOutput for two DP ranks
        # Rank 0 has 10 tokens, Rank 1 has 20 tokens
        output_0 = MagicMock(spec=SchedulerOutput)
        output_0.num_scheduled_tokens = {"req1": 10}
        output_0.total_num_scheduled_tokens = 10
        output_0.scheduled_new_reqs = []
        output_0.scheduled_spec_decode_tokens = {}
        output_0.scheduled_encoder_inputs = {}
        output_0.finished_req_ids = set()
        output_0.scheduled_cached_reqs = MagicMock(spec=CachedRequestData)
        output_0.scheduled_cached_reqs.req_ids = ["req1"]
        output_0.scheduled_cached_reqs.resumed_req_ids = set()
        output_0.scheduled_cached_reqs.new_token_ids = [[1, 2]]
        output_0.scheduled_cached_reqs.all_token_ids = {}
        output_0.scheduled_cached_reqs.new_block_ids = [None]
        output_0.scheduled_cached_reqs.num_computed_tokens = [10]
        output_0.scheduled_cached_reqs.num_output_tokens = [1]
        output_0.num_common_prefix_blocks = []

        output_1 = MagicMock(spec=SchedulerOutput)
        output_1.num_scheduled_tokens = {"req2": 20}
        output_1.total_num_scheduled_tokens = 20
        output_1.scheduled_new_reqs = []
        output_1.scheduled_spec_decode_tokens = {}
        output_1.scheduled_encoder_inputs = {}
        output_1.finished_req_ids = set()
        output_1.scheduled_cached_reqs = MagicMock(spec=CachedRequestData)
        output_1.scheduled_cached_reqs.req_ids = ["req2"]
        output_1.scheduled_cached_reqs.resumed_req_ids = set()
        output_1.scheduled_cached_reqs.new_token_ids = [[3, 4]]
        output_1.scheduled_cached_reqs.all_token_ids = {}
        output_1.scheduled_cached_reqs.new_block_ids = [None]
        output_1.scheduled_cached_reqs.num_computed_tokens = [20]
        output_1.scheduled_cached_reqs.num_output_tokens = [2]
        output_1.num_common_prefix_blocks = []

        scheduler.assigned_dp_rank = {"req1": 0, "req2": 1}

        combined = scheduler._combine_scheduler_outputs([output_0, output_1])

        assert combined.total_num_scheduled_tokens == 30

    def test_shutdown(self, mock_vllm_config, mock_kv_cache_config,
                      mock_structured_output_manager):
        """Test shutdown sends SHUTDOWN command to all workers."""
        scheduler = self._create_scheduler(mock_vllm_config,
                                           mock_kv_cache_config,
                                           mock_structured_output_manager)

        scheduler._send_command = MagicMock()
        scheduler._get_result = MagicMock(return_value=None)

        mock_process_0 = MagicMock()
        mock_process_1 = MagicMock()
        mock_process_0.is_alive = MagicMock(return_value=True)
        mock_process_1.is_alive = MagicMock(return_value=True)
        scheduler.processes = [mock_process_0, mock_process_1]

        scheduler.shutdown()

        # Verify SHUTDOWN commands were sent
        scheduler._send_command.assert_any_call(0, SchedulerCommand.SHUTDOWN)
        scheduler._send_command.assert_any_call(1, SchedulerCommand.SHUTDOWN)

        # Verify processes were joined
        mock_process_0.join.assert_called()
        mock_process_1.join.assert_called()


class TestUpdateVllmConfigForDPScheduler:
    """Test the update_vllm_config_for_dp_scheduler function."""

    def test_update_config_with_dp_size_greater_than_one(self):
        """Test Config is updated when DP size > 1."""
        mock_config = MagicMock()
        mock_config.sharding_config.total_dp_size = 2
        mock_config.scheduler_config._original_scheduler_cls = None
        mock_config.scheduler_config.scheduler_cls = "vllm.v1.core.sched.scheduler.Scheduler"
        mock_config.scheduler_config.async_scheduling = False

        update_vllm_config_for_dp_scheduler(mock_config)

        # Verify config was updated
        assert mock_config.scheduler_config._original_scheduler_cls == Scheduler
        assert mock_config.scheduler_config.scheduler_cls == DPScheduler

    def test_update_config_with_dp_size_one(self):
        """Test that config is NOT updated when DP size == 1."""
        mock_config = MagicMock()
        mock_config.sharding_config.total_dp_size = 1
        original_scheduler_cls = "vllm.v1.core.sched.scheduler.Scheduler"
        mock_config.scheduler_config.scheduler_cls = original_scheduler_cls

        update_vllm_config_for_dp_scheduler(mock_config)

        # Verify config was NOT changed
        assert mock_config.scheduler_config.scheduler_cls == original_scheduler_cls


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from unittest import mock

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from vllm.config import (CacheConfig, DeviceConfig, ModelConfig,
                         ParallelConfig, SchedulerConfig, SpeculativeConfig,
                         VllmConfig)
from vllm.config.load import LoadConfig

from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.runner import utils as runner_utils
from tpu_inference.spec_decode.jax.eagle3 import Eagle3Proposer

# Use a real model dir for config, but we will mock model loading/execution
model_dir = "meta-llama/Llama-3.1-8B-Instruct"
eagle3_dir = "yuhuili/EAGLE3-LLaMA3.1-Instruct-8B"


def _create_proposer(
    method: str,
    num_speculative_tokens: int,
) -> Eagle3Proposer:
    model_config = ModelConfig(model=model_dir,
                               runner="generate",
                               max_model_len=8192,
                               seed=42)

    speculative_config = SpeculativeConfig(
        target_model_config=model_config,
        target_parallel_config=ParallelConfig(),
        model=eagle3_dir,
        method=method,
        num_speculative_tokens=num_speculative_tokens,
    )

    vllm_config = VllmConfig(model_config=model_config,
                             cache_config=CacheConfig(block_size=16),
                             speculative_config=speculative_config,
                             device_config=DeviceConfig(device="tpu"),
                             parallel_config=ParallelConfig(
                                 pipeline_parallel_size=1,
                                 tensor_parallel_size=1),
                             load_config=LoadConfig(),
                             scheduler_config=SchedulerConfig(
                                 max_num_batched_tokens=8192,
                                 max_num_seqs=128,
                                 max_model_len=model_config.max_model_len,
                                 is_encoder_decoder=False))

    # Mock the runner, as the proposer needs it for initialization
    mock_runner = mock.MagicMock()
    # Create a real mesh for testing sharding-related logic
    devices = np.array(jax.devices())
    mock_runner.mesh = jax.sharding.Mesh(devices, axis_names=('model', ))
    mock_runner.max_num_tokens = 8192
    mock_runner.max_model_len = 8192
    mock_runner.kv_cache_config.kv_cache_groups = [mock.MagicMock()]
    mock_runner.input_batch = mock.MagicMock()

    return Eagle3Proposer(vllm_config=vllm_config, runner=mock_runner)


def test_prepare_inputs():
    """
    Mirrors the GPU test for prepare_inputs, adapted for JAX.
    - cu_target_query_lens: [0, a, a + b, a + b + c]
    - num_rejected_tokens: [n1, n2, n3]
    - num_tokens_per_req: [a - n1, b - n2, c - n3]
    - cu_num_tokens: [0, a - n1, a + b - n1 - n2, a + b + c - n1 - n2 - n3]
    - token_indices: [0, ..., a - n1 - 1, a, ..., a + b - n2 - 1, ...]
    """
    proposer = _create_proposer("eagle3", 1)
    num_reqs = 3
    max_num_seqs = 128
    max_num_blocks_per_req = 10  # Mock value

    # Mock runner attributes
    proposer.runner.input_batch.num_reqs = num_reqs
    proposer.runner.num_tokens_paddings = runner_utils.get_token_paddings(
        min_token_size=16, max_token_size=1024, padding_gap=0)

    # Mocks required by _prepare_draft_inputs helper
    proposer.combine_hidden_states_fn = lambda state, h: h  # Mock passthrough
    proposer.state = None  # Mock state
    proposer.runner.input_batch.block_table = [mock.MagicMock()]
    # Mock the block table return value (2D array)
    (proposer.runner.input_batch.block_table[0].get_cpu_tensor.return_value
     ) = jnp.zeros((num_reqs, max_num_blocks_per_req), dtype=jnp.int32)

    # --- Setup sequence data ---
    qsl_cpu = np.zeros(max_num_seqs + 1, dtype=np.int32)
    query_lens = np.zeros(max_num_seqs, dtype=np.int32)
    query_lens[:num_reqs] = [4, 7, 5]
    qsl_cpu[1:] = np.cumsum(query_lens)

    sl_cpu = np.zeros(max_num_seqs, dtype=np.int32)
    sl_cpu[:num_reqs] = [4, 7, 5]

    # Inputs
    total_tokens = 16
    hidden_size = 128
    # The input_ids should be large enough to be indexed by token_indices,
    # which can access up to total_tokens for padded requests.
    input_ids = jnp.arange(total_tokens + 1)
    aux_hidden_states = (jnp.ones((total_tokens + 1, hidden_size)),
                         jnp.ones((total_tokens + 1, hidden_size)),
                         jnp.ones((total_tokens + 1, hidden_size)))

    num_rejected_tokens_cpu = np.zeros(max_num_seqs, dtype=np.int32)
    num_rejected_tokens_cpu[:num_reqs] = [1, 3, 2]
    num_rejected_tokens = jnp.array(num_rejected_tokens_cpu)
    # This is only used in the _prepare_input_ids helper
    # It must be padded to max_num_seqs (128) to match the mask in jnp.where
    next_token_ids_cpu = np.zeros(max_num_seqs, dtype=np.int32)
    next_token_ids_cpu[:num_reqs] = [1, 2, 3]  # Valid tokens for active reqs
    next_token_ids = jnp.array(next_token_ids_cpu)

    attn_metadata = AttentionMetadata(
        seq_lens=jnp.array(sl_cpu),
        input_positions=jnp.arange(total_tokens),
        query_start_loc=jnp.array(qsl_cpu),
        block_tables=jnp.array([]),  # This will be replaced by the mock
        request_distribution=None,
    )
    attn_metadata.query_start_loc_cpu = qsl_cpu
    attn_metadata.seq_lens_cpu = sl_cpu

    # Expected results
    expected_new_qsl = np.zeros(max_num_seqs + 1, dtype=np.int32)
    num_tokens_per_req = np.zeros(max_num_seqs, dtype=np.int32)
    num_tokens_per_req[:num_reqs] = [3, 4, 3]
    # The implementation sets padded query lengths to 1, and rejected tokens
    # are 0 for padded requests.
    num_tokens_per_req[num_reqs:] = 1
    expected_new_qsl[1:] = np.cumsum(num_tokens_per_req)

    expected_new_seq_lens = np.zeros(max_num_seqs, dtype=np.int32)
    expected_new_seq_lens[:num_reqs] = [3, 4, 3]

    expected_total_tokens = int(expected_new_qsl[-1])
    expected_total_tokens = runner_utils.get_padded_token_len(
        proposer.runner.num_tokens_paddings, expected_total_tokens)

    expected_last_token_indices = jnp.array(expected_new_qsl[1:] - 1)

    # Execute
    target_hidden_states, input_ids, last_token_indices, updated_metadata = (
        proposer.prepare_inputs(attn_metadata, input_ids, aux_hidden_states,
                                next_token_ids, num_rejected_tokens))

    # Assertions
    assert jnp.array_equal(updated_metadata.query_start_loc,
                           jnp.array(expected_new_qsl))
    assert jnp.array_equal(updated_metadata.seq_lens,
                           jnp.array(expected_new_seq_lens))

    assert jnp.array_equal(last_token_indices, expected_last_token_indices)

    assert input_ids.shape == (expected_total_tokens, )
    # NOTE: We don't check the content of target_token_ids for padded requests
    # as it's complicated to construct the expected tensor. The shape check
    # and the qsl/seq_len checks are sufficient to validate the logic.
    # The concatenated hidden state shape should be (..., hidden_size * 3)
    assert target_hidden_states.shape == (expected_total_tokens,
                                          hidden_size * 3)


@pytest.mark.parametrize("method", ["eagle3"])
@pytest.mark.parametrize("num_speculative_tokens", [1, 3, 8])
def test_propose(method, num_speculative_tokens):
    proposer = _create_proposer(method, num_speculative_tokens)

    # Mock the JAX model functions
    hidden_size = 128
    vocab_size = 100
    batch_size = 2
    seq_len_1 = 5
    seq_len_2 = 3
    total_tokens = seq_len_1 + seq_len_2
    base_token_ids = [42, 60]

    def mock_model_fn(state, kv_caches, input_ids, target_hidden_states,
                      attn_metadata):
        """
        Mock model_fn.
        Returns: (kv_caches, hidden_states_for_logits, residual_tuple)

        - On first call (num_tokens == total_tokens):
          Populate hidden_states_for_logits[last_token_indices] with base_token_ids.
          Populate residual_tuple[0][last_token_indices] with base_token_ids.
        - On loop calls (num_tokens == batch_size):
          Use input_ids (previous draft token) to generate new token (input_ids + 1).
          Populate hidden_states_for_logits with (input_ids + 1).
          Populate residual_tuple[0] with (input_ids + 1).
        """
        num_tokens = input_ids.shape[0]

        # This will be used for logits (output 2)
        hidden_states_for_logits = jnp.zeros((num_tokens, hidden_size))
        # This will be fed into the next step (output 3, item 0)
        residual_hidden_states = jnp.zeros((num_tokens, hidden_size))

        if num_tokens == total_tokens:
            # First call in propose.
            # `propose` will select from last_token_indices.
            last_token_indices = attn_metadata.query_start_loc[1:] - 1

            # Set logits output
            hidden_states_for_logits = hidden_states_for_logits.at[
                last_token_indices, 0].set(jnp.array(base_token_ids))

            # Set residual for next step
            residual_hidden_states = residual_hidden_states.at[
                last_token_indices, 0].set(jnp.array(base_token_ids))
        else:
            # Subsequent calls in the loop
            # input_ids is the previous draft token (shape `batch_size`)
            # Mock logic: next token = previous token + 1
            next_token_ids_encoded = input_ids + 1

            # Set logits output
            hidden_states_for_logits = hidden_states_for_logits.at[:, 0].set(
                next_token_ids_encoded)

            # Set residual for next step
            residual_hidden_states = residual_hidden_states.at[:, 0].set(
                next_token_ids_encoded)

        # Return (kv_caches, hidden_states, residual_tuple)
        return kv_caches, hidden_states_for_logits, (residual_hidden_states, )

    def mock_compute_logits_fn(state, hidden_states, lora_metadata):
        # Create deterministic logits from hidden_states.
        # Takes the value from hidden_states[:, 0]
        token_ids = hidden_states[:, 0].astype(jnp.int32)
        return jax.nn.one_hot(token_ids, vocab_size)

    def mock_combine_hidden_states_fn(state, hidden_states):
        # Passthrough, as the mock doesn't need combination.
        return hidden_states

    proposer.model_fn = mock_model_fn
    proposer.compute_logits_fn = mock_compute_logits_fn
    proposer.combine_hidden_states_fn = mock_combine_hidden_states_fn
    proposer.state = None  # Mock state

    # Inputs
    kv_caches = [None] * 1  # Mock kv_caches

    # Create the 2D table first, as this is what the (unused) mock expects
    block_tables_2d = jnp.zeros((batch_size, 10), dtype=jnp.int32)

    attn_metadata = AttentionMetadata(
        seq_lens=jnp.array([seq_len_1, seq_len_2]),
        input_positions=jnp.concatenate(
            [jnp.arange(seq_len_1),
             jnp.arange(seq_len_2)]),
        query_start_loc=jnp.array([0, seq_len_1, total_tokens]),
        # Pass the FLATTENED table to simulate output of prepare_inputs
        block_tables=block_tables_2d.reshape(-1),
        request_distribution=None,
    )

    # These are the inputs to `propose`
    # input_ids (from prepare_inputs)
    target_token_ids = jnp.zeros(total_tokens, dtype=jnp.int32)
    # target_hidden_states (from prepare_inputs)
    target_hidden_states = jnp.zeros((total_tokens, hidden_size))
    # last_token_indices (from prepare_inputs)
    last_token_indices = attn_metadata.query_start_loc[1:] - 1

    # Mock runner for block tables
    # This mock isn't actually used by propose(), but we'll set it
    # to the 2D table for correctness, as that's what
    # _prepare_draft_inputs (called by prepare_inputs) would expect.
    proposer.runner.input_batch.num_reqs = batch_size
    proposer.runner.input_batch.block_table = [mock.MagicMock()]
    (proposer.runner.input_batch.block_table[0].get_device_tensor.return_value
     ) = block_tables_2d

    # Execute
    _, draft_token_ids = proposer.propose(
        kv_caches,
        target_token_ids,
        attn_metadata,
        last_token_indices,
        target_hidden_states,
    )

    if draft_token_ids.ndim == 1:
        draft_token_ids = jnp.expand_dims(draft_token_ids, axis=-1)
    # Assertions
    assert draft_token_ids.shape == (batch_size, num_speculative_tokens)

    # Check the generated tokens
    # Step 0: base_token_ids [42, 60]
    # Step 1: [43, 61]
    # Step 2: [44, 62]
    # ...
    expected_tokens = np.zeros((batch_size, num_speculative_tokens),
                               dtype=np.int64)
    for i in range(batch_size):
        for j in range(num_speculative_tokens):
            expected_tokens[i, j] = base_token_ids[i] + j

    assert jnp.array_equal(draft_token_ids, jnp.array(expected_tokens))



# SPDX-License-Identifier: Apache-2.0

import os

import jax
import jax.numpy as jnp
from absl.testing import absltest, parameterized
from jax._src import test_util as jtu

from tpu_inference import utils
from tpu_inference.kernels.collectives import all_gather_matmul

jax.config.parse_flags_with_absl()

P = jax.sharding.PartitionSpec

SpongeDir: str | None = os.environ.get('TEST_UNDECLARED_OUTPUTS_DIR', None)


@jtu.with_config(jax_numpy_dtype_promotion='standard')
class AllGatherMatmulTest(jtu.JaxTestCase):

    @parameterized.product(
        grid_k=[1, 2, 3],
        grid_n=[1, 2, 3],
        rhs_transpose=[True, False],
    )
    def test_all_gather_matmul(self, grid_k, grid_n, rhs_transpose):
        if jax.device_count() != 8:
            self.skipTest('Not enough devices for test')

        axis_name = 'x'
        num_devices = jax.device_count()
        mesh = utils.make_optimized_mesh((num_devices, ), (axis_name, ))
        bk, bn = 1024, 1024
        m, k, n = 1024, bk * grid_k, bn * grid_n * num_devices

        # Run the test 10 times to expose race conditions as much as possible.
        for i in range(10):
            # Create input data
            prng_key = jax.random.key(1234 + i)
            k0, k1 = jax.random.split(prng_key, 2)
            x = jax.random.normal(k0, (m, k), dtype=jnp.bfloat16)
            y_shape = (n, k) if rhs_transpose else (k, n)
            y_sharding = P(axis_name, None) if rhs_transpose else P(
                None, axis_name)
            y = jax.random.normal(k1, y_shape, dtype=jnp.bfloat16)
            sharded_x = jax.device_put(
                x, jax.sharding.NamedSharding(mesh, P(axis_name, None)))
            sharded_y = jax.device_put(
                y, jax.sharding.NamedSharding(mesh, y_sharding))

            # Run the all_gather_matmul function
            output = all_gather_matmul.all_gather_matmul(
                sharded_x,
                sharded_y,
                mesh,
                axis_name,
                bk=bk,
                bn=bn,
                rhs_transpose=rhs_transpose,
            )
            y_for_dot = sharded_y.T if rhs_transpose else sharded_y
            expected_output = jnp.dot(sharded_x, y_for_dot)
            self.assertAllClose(output, expected_output, atol=1e-2, rtol=1e-2)


if __name__ == "__main__":
    absltest.main(testLoader=jtu.JaxTestLoader())



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Register function to overwrite torch functions used by vllm via torchax."""

import jax
import jax.numpy as jnp

from tpu_inference.layers.common.attention_interface import \
    sharded_flash_attention


def scaled_dot_product_attention(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    scale=None,
    enable_gqa=False,
    *,
    mesh: jax.sharding.Mesh,
):
    """The same args as torch.nn.functional.scaled_dot_product_attention to use flash attention."""
    if dropout_p != 0.0:
        raise NotImplementedError("patched_sdpa does not support dropout_p")
    if enable_gqa is not False:
        raise NotImplementedError("patched_sdpa does not support enable_gqa")

    # Q, K, V shapes: (batch, num_heads, seq_len, head_dim)
    batch = query.shape[0]
    num_heads = query.shape[1]
    q_seq_len = query.shape[2]
    kv_seq_len = key.shape[2]

    # padding due to the requirement of sharded_flash_attention
    q_pad = (128 - (q_seq_len % 128)) % 128
    kv_pad = (128 - (kv_seq_len % 128)) % 128

    if q_pad > 0:
        query = jnp.pad(query, ((0, 0), (0, 0), (0, q_pad), (0, 0)))
    if kv_pad > 0:
        key = jnp.pad(key, ((0, 0), (0, 0), (0, kv_pad), (0, 0)))
        value = jnp.pad(value, ((0, 0), (0, 0), (0, kv_pad), (0, 0)))

    # Prevent nan while using -inf
    mask_value = -0.7 * float(jnp.finfo(jnp.dtype("float32")).max)
    attention_bias = jnp.zeros((batch, num_heads, q_seq_len, kv_seq_len),
                               dtype=jnp.float32)
    if attn_mask is not None:
        # attn_mask shape: (batch, num_heads, q_len, kv_len)
        if attn_mask.dtype == jnp.bool_:
            attention_bias = jnp.where(attn_mask, attention_bias, mask_value)
        else:
            attention_bias += attn_mask

    if q_pad > 0 or kv_pad > 0:
        attention_bias = jnp.pad(
            attention_bias,
            ((0, 0), (0, 0), (0, q_pad), (0, kv_pad)),
            mode="constant",
            constant_values=mask_value,
        )

    attn_fn = sharded_flash_attention(mesh,
                                      causal=is_causal,
                                      sm_scale=scale,
                                      use_attention_bias=True)
    out = attn_fn(query, key, value, attention_bias, None)

    if q_pad > 0:
        out = out[:, :, :q_seq_len, :]

    return out
